/*
Copyright (c) 2009, Joel Reardon
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the University of Waterloo nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#include "spe.h"

#include <dlfcn.h>
#include <sys/file.h>
#include <string.h>


#define SPE_FL_MISSING -1
#define SPE_FL_LOADED 1
#define SPE_FL_UNLOADED 0 



extern const char* SYMBOLS[];
// N.B. this has to be defined by the user. It consists of a NULL termining list of symbol names that are 
// implemented in the experimental callback library. eg: {"buffer", "connection", NULL}



spe_fl_t * spe_fl_new() {
	int n_symbols;
	const char** p_symbols = SYMBOLS;
	while (*p_symbols) {
		++p_symbols;
		++n_symbols;
	}
	size_t array_size = sizeof(int)*n_symbols;
	spe_fl_t * fl = malloc(sizeof(spe_fl_t));
	fl->lock = malloc(sizeof(pthread_mutex_t));	
	
	fl->symbol_state = malloc(array_size);
	fl->symbol_link = malloc(array_size);
	memset(fl->symbol_state, 0, array_size);
	memset(fl->symbol_link, 0, array_size);
	pthread_mutex_init(fl->lock, NULL);
	
	fl->dlhandle = NULL;
	return fl;

}

void spe_fl_delete(spe_fl_t* fl)  {
	if (fl->dlhandle) { //unloader has ran before, wait on thread if needed
		pthread_join(fl->unloader, NULL);
		pthread_create(&(fl->unloader), NULL, spe_fl_unload, fl->dlhandle); 
		pthread_join(fl->unloader, NULL);
	}
	pthread_mutex_destroy(fl->lock);
	free(fl->lock);
	free(fl->symbol_state);
	free(fl->symbol_link);
	
	free(fl);
	
}

spe_obs_func spe_fl_get(spe_fl_t *fl, int family) {
	void* fn;
	pthread_mutex_lock(fl->lock);
	if (fl->symbol_state[family] == SPE_FL_MISSING) {
		fn = (void*)NULL;
	} 
	else {
		fn = (void*)fl->symbol_link[family];
		if (!fn) {
			fn = spe_fl_set(fl, family);
		}
	}
	pthread_mutex_unlock(fl->lock);
	return fn;
}

void spe_fl_insert(spe_fl_t * fl, int family, void* fn) {
	if (!fn) {
		fl->symbol_state[family] = SPE_FL_MISSING;
	}
	else {
		fl->symbol_state[family] = SPE_FL_LOADED;
		fl->symbol_link[family] = (int)fn;
	}
}

void* spe_fl_set(spe_fl_t *fl, int family) {
	char * dle;
	void * fn;
	dlerror();
	fn = dlsym(fl->dlhandle, SYMBOLS[family]);
	dle = dlerror();
	spe_fl_insert(fl, family, fn);
	if (!dle) {
		;//write to a log file
	}
	return fn;
}


//call only from spe_fl_load
void spe_fl_resym_all(spe_fl_t* fl) {
	int i = 0; 
	const char** p_symbol = SYMBOLS;
	for (; *p_symbol; ++i, ++p_symbol) {
		if (fl->symbol_state[i]) {
			spe_fl_set(fl, i);	
		}		
	}
}


//call only from spe_fl_load
void* spe_fl_unload(void* handle) {
	if (handle) {
		struct timeval tv;
		tv.tv_usec = 0;
		tv.tv_sec = 10;
		select(0, NULL,NULL,NULL,&tv);
		dlclose(handle);
	}
	pthread_exit(NULL);
}

char* spe_fl_load(spe_fl_t * fl, char* library) {

	void* handle;
	void* oldh;
	char* err;
	
	if (!fl) return;
	pthread_mutex_lock(fl->lock);

	handle = dlopen(library, RTLD_LAZY);
	
	oldh = fl->dlhandle;
	err = dlerror();
	if (err) {
		handle = NULL;
		pthread_mutex_unlock(fl->lock);
		return err;
	}
	if (oldh) //unloader has ran before, wait on thread if needed
		pthread_join(fl->unloader, NULL);
	pthread_create(&(fl->unloader), NULL, spe_fl_unload, oldh); 
	fl->dlhandle = handle;
	spe_fl_resym_all(fl);
	
	pthread_mutex_unlock(fl->lock);
	return NULL;
}
